from __future__ import annotations

import base64
import json
import logging
from decimal import Decimal
from io import BytesIO
from typing import TYPE_CHECKING
from typing import TypedDict

import qrcode
from django.core.paginator import Paginator
from django_renderpdf.helpers import render_pdf

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import IO

    from PIL.Image import Image

    from django_afip.models import Receipt
    from django_afip.models import ReceiptEntry
    from django_afip.models import ReceiptPDF


logger = logging.getLogger(__name__)


class ReceiptQrCode:
    """A QR code for receipt

    See: https://www.afip.gob.ar/fe/qr/especificaciones.asp"""

    BASE_URL = "https://www.afip.gob.ar/fe/qr/?p="

    def __init__(self, receipt: Receipt) -> None:
        self._receipt = receipt
        # The examples on the website say that "importe" and "ctz" are both "decimal"
        # type. JS/JSON has no decimal type. The examples use integeres.
        #
        # Using integers would drop cents, which would result in mismatching
        # information. Using strings would add quotes, which likely breaks.
        #
        # Using floats seems to be the only viable solution, and SHOULD be fine for
        # values in the range supported.
        self._data = {
            "ver": 1,
            "fecha": receipt.issued_date.strftime("%Y-%m-%d"),
            "cuit": receipt.point_of_sales.owner.cuit,
            "ptoVta": receipt.point_of_sales.number,
            "tipoCmp": int(receipt.receipt_type.code),
            "nroCmp": receipt.receipt_number,
            "importe": float(receipt.total_amount),
            "moneda": receipt.currency.code,
            "ctz": float(receipt.currency_quote),
            "tipoDocRec": int(receipt.document_type.code),
            "nroDocRec": receipt.document_number,
            "tipoCodAut": "E",  # TODO: need to implement CAEA
            "codAut": int(receipt.validation.cae),
        }

    def as_png(self) -> Image:
        json_data = json.dumps(self._data)
        encoded_json = base64.b64encode(json_data.encode()).decode()
        url = f"{self.BASE_URL}{encoded_json}"

        qr = qrcode.QRCode(version=1, border=4)
        qr.add_data(url)
        qr.make()

        return qr.make_image()


def get_encoded_qrcode(receipt_pdf: ReceiptPDF) -> str:
    """Return a QRCode encoded for embeding in HTML."""

    img_data = BytesIO()
    qr_img = ReceiptQrCode(receipt_pdf.receipt).as_png()
    qr_img.save(img_data, format="PNG")

    return base64.b64encode(img_data.getvalue()).decode()


# Note: When updating this, be sure to update the docstring of the method that uses
# these below.
TEMPLATE_NAMES = [
    "receipts/{taxpayer}/pos_{point_of_sales}/code_{code}.html",
    "receipts/{taxpayer}/code_{code}.html",
    "receipts/code_{code}.html",
    "receipts/{code}.html",
]


class EntriesForPage(TypedDict):
    previous_subtotal: Decimal
    subtotal: Decimal
    entries: Iterable[ReceiptEntry]


def create_entries_context_for_render(
    paginator: Paginator,
) -> dict[int, EntriesForPage]:
    entries: dict[int, EntriesForPage] = {}
    subtotal = Decimal(0)
    for i in paginator.page_range:
        previous_subtotal = subtotal
        page = paginator.get_page(i)

        for entry in page.object_list:
            subtotal += entry.total_price

        entries[i] = {
            "previous_subtotal": previous_subtotal,
            "subtotal": subtotal,
            "entries": paginator.get_page(i).object_list,
        }
    return entries


class PdfBuilder:
    """Builds PDF files for Receipts.

    Creating a new instance of a builder does nothing; use :meth:`~render_pdf` to
    actually render the file.

    This type can be subclassed to add custom behaviour or data into PDF files.
    """

    def __init__(self, entries_per_page: int = 15) -> None:
        self.entries_per_page = entries_per_page

    def get_template_names(self, receipt: Receipt) -> list[str]:
        """Return the templates use to render the Receipt PDF.

        Template discovery tries to find any of the below receipts::

            receipts/{taxpayer}/pos_{point_of_sales}/code_{code}.html
            receipts/{taxpayer}/code_{code}.html
            receipts/code_{code}.html
            receipts/{code}.html

        To override, for example, the "Factura C" template for point of sales 0002 for
        Taxpayer 20-32964233-0, use::

            receipts/20329642330/pos_2/code_6.html
        """
        return [
            template.format(
                taxpayer=receipt.point_of_sales.owner.cuit,
                point_of_sales=receipt.point_of_sales.number,
                code=receipt.receipt_type.code,
            )
            for template in TEMPLATE_NAMES
        ]

    def get_context(self, receipt: Receipt) -> dict:
        """Returns the context used to render the PDF file."""
        from django_afip.models import Receipt
        from django_afip.models import ReceiptPDF

        context: dict = {}

        receipt_pdf = (
            ReceiptPDF.objects.select_related(
                "receipt",
                "receipt__receipt_type",
                "receipt__document_type",
                "receipt__validation",
                "receipt__point_of_sales",
                "receipt__point_of_sales__owner",
            )
            .prefetch_related(
                "receipt__entries",
            )
            .get(receipt=receipt)
        )

        # Prefetch required data in a single query:
        receipt_pdf.receipt = (
            Receipt.objects.select_related(
                "receipt_type",
                "document_type",
                "validation",
                "point_of_sales",
                "point_of_sales__owner",
            )
            .prefetch_related(
                "entries",
            )
            .get(
                pk=receipt_pdf.receipt_id,
            )
        )
        taxpayer = receipt_pdf.receipt.point_of_sales.owner
        entries = receipt_pdf.receipt.entries.order_by("id")
        paginator = Paginator(entries, self.entries_per_page)

        context["entries"] = create_entries_context_for_render(paginator)
        context["pdf"] = receipt_pdf
        context["taxpayer"] = taxpayer
        context["qrcode"] = get_encoded_qrcode(receipt_pdf)

        return context

    def render_pdf(self, receipt: Receipt, file_: IO) -> None:
        """Renders the PDF into ``file_``."""
        render_pdf(
            template=self.get_template_names(receipt),
            file_=file_,
            context=self.get_context(receipt),
        )



from __future__ import annotations


class DjangoAfipException(Exception):
    """Superclass for exceptions thrown by `django_afip`."""


class AfipException(DjangoAfipException):
    """Wraps around errors returned by AFIP's WS."""

    def __init__(self, response) -> None:  # noqa: ANN001
        if "Errors" in response:
            message = (
                f"Error {response.Errors.Err[0].Code}: {response.Errors.Err[0].Msg}"
            )
        else:
            message = (
                f"Error {response.errorConstancia.idPersona}: "
                f"{response.errorConstancia.error[0]}"
            )
        Exception.__init__(self, message)


class AuthenticationError(DjangoAfipException):
    """Raised when there is an error during an authentication attempt.

    Usually, subclasses of this error are raised, but for unexpected errors, this type
    may be raised.
    """


class CertificateExpired(AuthenticationError):
    """Raised when an authentication was attempted with an expired certificate."""


class UntrustedCertificate(AuthenticationError):
    """Raise when an untrusted certificate is used in an authentication attempt."""


class CorruptCertificate(AuthenticationError):
    """Raised when a corrupt certificate file is used in an authentication attempt."""


class CannotValidateTogether(DjangoAfipException):
    """Raised when attempting to validate invalid combinations of receipts.

    Receipts of different ``receipt_type`` or ``point_of_sales`` cannot be validated
    together.
    """


class ValidationError(DjangoAfipException):
    """Raised when a single Receipt failed to validate with AFIP's WS."""



from __future__ import annotations

import random
import re
from datetime import date
from unittest.mock import patch

import pytest
from django.core.paginator import Paginator

from django_afip import factories
from django_afip import models
from django_afip.pdf import PdfBuilder
from django_afip.pdf import ReceiptQrCode
from django_afip.pdf import create_entries_context_for_render


@pytest.mark.django_db
def test_pdf_generation() -> None:
    """Test PDF file generation.

    For the moment, this test case mostly verifies that pdf generation
    *works*, but does not actually validate the pdf file itself.

    Running this locally *will* yield the file itself, which is useful for
    manual inspection.
    """
    pdf = factories.ReceiptPDFFactory(receipt__receipt_number=3)
    factories.ReceiptValidationFactory(receipt=pdf.receipt)
    pdf.save_pdf()
    regex = r"afip/receipts/[a-f0-9]{2}/[a-f0-9]{2}/[a-f0-9]{32}.pdf"

    assert re.match(regex, pdf.pdf_file.name)
    assert pdf.pdf_file.name.endswith(".pdf")


@pytest.mark.django_db
def test_unauthorized_receipt_generation() -> None:
    """
    Test PDF file generation for unauthorized receipts.

    Confirm that attempting to generate a PDF for an unauthorized receipt
    raises.
    """
    taxpayer = factories.TaxPayerFactory()
    receipt = factories.ReceiptFactory(
        receipt_number=None,
        point_of_sales__owner=taxpayer,
    )
    pdf = models.ReceiptPDF.objects.create_for_receipt(
        receipt=receipt,
        client_name="John Doe",
        client_address="12 Green Road\nGreenville\nUK",
    )
    with pytest.raises(
        Exception, match="Cannot generate pdf for non-authorized receipt"
    ):
        pdf.save_pdf()


@pytest.mark.django_db
def test_signal_generation_for_not_validated_receipt() -> None:
    printable = factories.ReceiptPDFFactory()

    assert not (printable.pdf_file)


@pytest.mark.django_db
def test_qrcode_data() -> None:
    pdf = factories.ReceiptPDFFactory(
        receipt__receipt_number=3,
        receipt__issued_date=date(2021, 3, 2),
    )
    factories.ReceiptValidationFactory(receipt=pdf.receipt)

    qrcode = ReceiptQrCode(pdf.receipt)
    assert qrcode._data == {
        "codAut": 67190616790549,
        "ctz": 1.0,
        "cuit": 20329642330,
        "fecha": "2021-03-02",
        "importe": 130.0,
        "moneda": "PES",
        "nroCmp": 3,
        "nroDocRec": 203012345,
        "ptoVta": 1,
        "tipoCmp": 6,
        "tipoCodAut": "E",
        "tipoDocRec": 96,
        "ver": 1,
    }


@pytest.mark.django_db
def test_create_entries_for_render() -> None:
    validation = factories.ReceiptValidationFactory()
    for _i in range(10):
        factories.ReceiptEntryFactory(
            receipt=validation.receipt, unit_price=1, quantity=1
        )
    entries_queryset = models.ReceiptEntry.objects.all()
    paginator = Paginator(entries_queryset, 5)
    entries = create_entries_context_for_render(paginator)

    assert list(entries.keys()) == [1, 2]
    assert entries[1]["previous_subtotal"] == 0
    assert entries[1]["subtotal"] == 5
    assert list(entries[1]["entries"]) == list(models.ReceiptEntry.objects.all()[:5])

    assert entries[2]["previous_subtotal"] == 5
    assert entries[2]["subtotal"] == 10
    assert list(entries[2]["entries"]) == list(models.ReceiptEntry.objects.all()[5:10])


@pytest.mark.django_db
def test_receipt_pdf_modified_builder() -> None:
    validation = factories.ReceiptValidationFactory()
    validation.receipt.total_amount = 20
    validation.receipt.save()
    for _i in range(10):
        random.uniform(1.00, 12.5)
        factories.ReceiptEntryFactory(
            receipt=validation.receipt, unit_price=1, quantity=2
        )

    printable = factories.ReceiptPDFFactory(receipt=validation.receipt)
    assert not printable.pdf_file

    printable.save_pdf(builder=PdfBuilder(entries_per_page=5))
    assert printable.pdf_file
    assert printable.pdf_file.name.endswith(".pdf")


@pytest.mark.django_db
def test_receipt_pdf_call_function() -> None:
    validation = factories.ReceiptValidationFactory()
    for _i in range(80):
        price = random.uniform(1.00, 12.5)
        factories.ReceiptEntryFactory(
            receipt=validation.receipt, unit_price=price, quantity=2
        )

    printable = factories.ReceiptPDFFactory(receipt=validation.receipt)
    with patch(
        "django_afip.pdf.create_entries_context_for_render", spec=True
    ) as mocked_call:
        printable.save_pdf(builder=PdfBuilder(entries_per_page=5))

    assert mocked_call.called
    assert mocked_call.call_count == 1



from __future__ import annotations

import pytest
from django.core import management

from django_afip.models import GenericAfipType


@pytest.mark.django_db
def test_afip_metadata_command() -> None:
    assert len(GenericAfipType.SUBCLASSES) == 7

    for model in GenericAfipType.SUBCLASSES:
        # TYPING: mypy doesn't know about `.objects`.
        assert model.objects.count() == 0  # type: ignore[attr-defined]

    management.call_command("afipmetadata")

    for model in GenericAfipType.SUBCLASSES:
        # TYPING: mypy doesn't know about `.objects`.
        assert model.objects.count() > 0  # type: ignore[attr-defined]



from __future__ import annotations

from unittest.mock import patch

import pytest
import requests
from requests.exceptions import SSLError

from django_afip.clients import get_client


@pytest.mark.live
def test_services_are_cached() -> None:
    service1 = get_client("wsfe", False)
    with patch.dict("django_afip.clients.WSDLS", values={}, clear=True):
        service2 = get_client("wsfe", False)

    assert service1 is service2


def test_inexisting_service() -> None:
    with pytest.raises(ValueError, match="Unknown service name, nonexistant"):
        get_client("nonexistant", False)


@pytest.mark.live
def test_insecure_dh_hack_required() -> None:
    with pytest.raises(SSLError, match="SSL: DH_KEY_TOO_SMALL. dh key too small"):
        requests.get("https://servicios1.afip.gov.ar/wsfev1/service.asmx?WSDL")



from __future__ import annotations

from typing import TYPE_CHECKING

from django.db.models.signals import post_save
from django.db.models.signals import pre_save
from django.dispatch import receiver

from django_afip import models

if TYPE_CHECKING:
    from django.db.models import Model


@receiver(pre_save, sender=models.TaxPayer)
def update_certificate_expiration(
    sender: type[Model],
    instance: models.TaxPayer,
    **kwargs,
) -> None:
    if instance.certificate:
        instance.certificate_expiration = instance.get_certificate_expiration()


FILE_FIELDS = ["certificate", "logo"]


# Store old files before saving.
@receiver(pre_save, sender=models.TaxPayer)
def store_old_files(
    sender: type[models.TaxPayer],
    instance: models.TaxPayer,
    **kwargs,
) -> None:
    if instance.pk:
        try:
            old_instance = sender.objects.get(pk=instance.pk)
            # Save the reference of the old files in the model.
            instance._old_files = {
                field: getattr(old_instance, field) for field in FILE_FIELDS
            }
        except sender.DoesNotExist:
            instance._old_files = {}


# Delete old files after saving.
@receiver(post_save, sender=models.TaxPayer)
def delete_file_taxpayer(
    sender: type[models.TaxPayer],
    instance: models.TaxPayer,
    **kwargs,
) -> None:
    if not instance.pk:
        return  # The instance is new, there are no old files to delete.

    old_files = getattr(instance, "_old_files", {})

    for field_name in FILE_FIELDS:
        old_file = old_files.get(field_name)
        new_file = getattr(instance, field_name)

        if old_file and old_file != new_file:
            # Delete the old file from storage.
            old_file.delete(save=False)



# Generated by Django 5.1 on 2024-08-08 22:04
from __future__ import annotations

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("afip", "0013_alter_receiptentry_quantity"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pointofsales",
            name="blocked",
            field=models.BooleanField(default=False, verbose_name="blocked"),
        ),
    ]



from __future__ import annotations

import logging
from contextlib import contextmanager
from datetime import datetime
from typing import TYPE_CHECKING
from typing import TypedDict

from django.contrib import admin
from django.contrib import messages
from django.db.models import F
from django.db.models import QuerySet
from django.http import HttpRequest
from django.http import HttpResponse
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext as _

from django_afip import exceptions
from django_afip import models
from django_afip.models import ReceiptQuerySet

if TYPE_CHECKING:
    from collections.abc import Generator
    from collections.abc import Iterable
    from collections.abc import Sequence

    from django_stubs_ext import WithAnnotations

logger = logging.getLogger(__name__)


# TODO: Add an action to populate generic types.


@contextmanager
def catch_errors(
    self: admin.ModelAdmin,
    request: HttpRequest,
) -> Generator[None, None, None]:
    """Catches specific errors in admin actions and shows a friendly error."""
    try:
        yield
    except exceptions.CertificateExpired as e:
        logger.exception(e)
        self.message_user(
            request,
            _("The AFIP Taxpayer certificate has expired."),
            messages.ERROR,
        )
    except exceptions.UntrustedCertificate as e:
        logger.exception(e)
        self.message_user(
            request,
            _("The AFIP Taxpayer certificate is untrusted."),
            messages.ERROR,
        )
    except exceptions.CorruptCertificate as e:
        logger.exception(e)
        self.message_user(
            request,
            _("The AFIP Taxpayer certificate is corrupt."),
            messages.ERROR,
        )
    except exceptions.AuthenticationError as e:
        logger.exception(e)
        self.message_user(
            request,
            _("An unknown authentication error has ocurred: %s") % e,
            messages.ERROR,
        )


class VatInline(admin.TabularInline):
    model = models.Vat
    fields = (
        "vat_type",
        "base_amount",
        "amount",
    )
    extra = 1


class TaxInline(admin.TabularInline):
    model = models.Tax
    fields = (
        "tax_type",
        "description",
        "base_amount",
        "aliquot",
        "amount",
    )
    extra = 1


class ReceiptEntryInline(admin.TabularInline):
    model = models.ReceiptEntry
    fields = (
        "description",
        "quantity",
        "unit_price",
        "vat",
    )
    extra = 1


class ReceiptValidationInline(admin.StackedInline):
    model = models.ReceiptValidation
    readonly_fields = (
        "result",
        "processed_date",
        "cae",
        "cae_expiration",
        "observations",
    )
    extra = 0


class ReceiptStatusFilter(admin.SimpleListFilter):
    title = _("status")
    parameter_name = "status"

    VALIDATED = "validated"
    NOT_VALIDATED = "not_validated"

    def lookups(
        self,
        request: HttpRequest,
        model_admin: admin.ModelAdmin,
    ) -> Iterable[tuple[str, str]]:
        return (
            (self.VALIDATED, _("Validated")),
            (self.NOT_VALIDATED, _("Not validated")),
        )

    # TODO: generics
    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet:
        if self.value() == self.VALIDATED:
            return queryset.filter(
                validation__result=models.ReceiptValidation.RESULT_APPROVED
            )
        if self.value() == self.NOT_VALIDATED:
            return queryset.exclude(
                validation__result=models.ReceiptValidation.RESULT_APPROVED
            )
        return queryset


class ReceiptTypeFilter(admin.SimpleListFilter):
    title = models.ReceiptType._meta.verbose_name
    parameter_name = "receipt_type"

    def lookups(
        self,
        request: HttpRequest,
        model_admin: admin.ModelAdmin,
    ) -> Iterable[tuple[str, str]]:
        return (
            (receipt_type.code, receipt_type.description)
            for receipt_type in models.ReceiptType.objects.filter(
                receipts__isnull=False,
            ).distinct()
        )

    def queryset(self, request: HttpRequest, queryset: QuerySet) -> QuerySet:
        value = self.value()
        if value:
            queryset = queryset.filter(receipt_type__code=value)
        return queryset


@admin.register(models.Receipt)
class ReceiptAdmin(admin.ModelAdmin):
    search_fields = ("receipt_number",)
    list_display = (
        "id",
        "receipt_type",
        "point_of_sales",
        "number",
        "issued_date",
        "friendly_total_amount",
        "validated",
        "pdf_link",
    )
    list_filter = (
        ReceiptStatusFilter,
        ReceiptTypeFilter,
    )
    autocomplete_fields = (
        "currency",
        "receipt_type",
        "related_receipts",
    )
    date_hierarchy = "issued_date"

    __related_fields = (
        "validated",
        "cae",
    )

    inlines = (
        VatInline,
        TaxInline,
        ReceiptEntryInline,
        ReceiptValidationInline,
    )
    ordering = ("-issued_date",)

    readonly_fields = __related_fields

    class Annotations(TypedDict):
        pdf_id: int
        validation_result: str

    def get_queryset(self, request: HttpRequest) -> QuerySet:
        return (
            super()
            .get_queryset(request)
            .select_related(
                "receipt_type",
                "point_of_sales",
            )
            .annotate(
                pdf_id=F("receiptpdf__id"),
                validation_result=F("validation__result"),
            )
        )

    @admin.display(description=_("receipt number"), ordering="receipt_number")
    def number(self, obj: models.Receipt) -> str | None:
        return obj.formatted_number

    @admin.display(description=_("total amount"))
    def friendly_total_amount(self, obj: models.Receipt) -> str:
        return "{:0.2f} ARS{}".format(
            obj.total_amount * obj.currency_quote,
            "*" if obj.currency_quote != 1 else "",
        )

    @admin.display(
        boolean=True,
        description=_("validated"),
        ordering="validation__result",
    )
    def validated(self, obj: WithAnnotations[models.Receipt, Annotations]) -> bool:
        validation_result = obj.validation_result
        return validation_result == models.ReceiptValidation.RESULT_APPROVED

    @admin.display(description=_("PDF"), ordering="receiptpdf__id")
    def pdf_link(self, obj: WithAnnotations[models.Receipt, Annotations]) -> str:
        if not obj.pdf_id:
            return format_html(
                '<a href="{}?receipt={}">{}</a>',
                reverse(self.admin_site.name + ":afip_receiptpdf_add"),
                obj.id,
                _("Create"),
            )
        return format_html(
            '<a href="{}">{}</a>',
            reverse(
                self.admin_site.name + ":afip_receiptpdf_change",
                args=(obj.pdf_id,),
            ),
            _("Edit"),
        )

    @admin.display(description=_("cae"), ordering="validation__cae")
    def cae(self, obj: models.Receipt) -> str:
        return obj.validation.cae

    @admin.action(description=_("Validate"))
    def validate(
        self,
        request: HttpRequest,
        queryset: QuerySet[models.Receipt],
    ) -> None:
        assert isinstance(queryset, ReceiptQuerySet)  # required for mypy
        with catch_errors(self, request):
            errs = queryset.validate()

        if errs:
            self.message_user(
                request,
                _("Receipt validation failed: %s.") % errs,
                messages.ERROR,
            )

    actions = (validate,)


@admin.register(models.AuthTicket)
class AuthTicketAdmin(admin.ModelAdmin):
    list_display = (
        "unique_id",
        "owner",
        "service",
        "generated",
        "expires",
    )


@admin.register(models.TaxPayer)
class TaxPayerAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "cuit",
        "certificate_expiration",
    )

    @admin.action(description=_("Fetch points of sales"))
    def fetch_points_of_sales(
        self,
        request: HttpRequest,
        queryset: QuerySet[models.TaxPayer],
    ) -> None:
        with catch_errors(self, request):
            poses = [
                pos
                for taxpayer in queryset.all()
                for pos in taxpayer.fetch_points_of_sales()
            ]

            created = len([pos for pos in poses if pos[1]])
            skipped = len(poses) - created

            self.message_user(
                request,
                message=(_("%d points of sales created.") % created),
                level=messages.SUCCESS,
            )
            self.message_user(
                request,
                message=(_("%d points of sales already existed.") % skipped),
                level=messages.WARNING,
            )

    @admin.action(description=_("Generate key"))
    def generate_key(
        self,
        request: HttpRequest,
        queryset: QuerySet[models.TaxPayer],
    ) -> None:
        key_count = sum(t.generate_key() for t in queryset.all())

        if key_count == 1:
            message = _("Key generated successfully.")
            level = messages.SUCCESS
        elif key_count:
            message = _("%d keys generated successfully.") % key_count
            level = messages.SUCCESS
        else:
            message = _("No keys generated; Taxpayers already had keys.")
            level = messages.ERROR

        self.message_user(
            request,
            message=message,
            level=level,
        )

    @admin.action(description=_("Generate CSR"))
    def generate_csr(
        self,
        request: HttpRequest,
        queryset: QuerySet[models.TaxPayer],
    ) -> HttpResponse | None:
        if queryset.count() > 1:
            self.message_user(
                request,
                message=_("Can only generate CSR for one taxpayer at a time."),
                level=messages.ERROR,
            )
            return None

        taxpayer = queryset.first()
        assert taxpayer is not None, "At least one taxpayer must be selected"
        if not taxpayer.key:
            taxpayer.generate_key()

        csr = taxpayer.generate_csr()
        filename = f"cuit-{taxpayer.cuit}-{int(datetime.now().timestamp())}.csr"

        response = HttpResponse(content_type="application/pkcs10")
        response["Content-Disposition"] = f"attachment; filename={filename}"

        response.write(csr.read())
        return response

    actions = (
        fetch_points_of_sales,
        generate_key,
        generate_csr,
    )


@admin.register(models.PointOfSales)
class PointOfSalesAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "owner",
        "number",
        "issuance_type",
        "blocked",
        "drop_date",
    )


@admin.register(models.CurrencyType)
class CurrencyTypeAdmin(admin.ModelAdmin):
    search_fields = (
        "code",
        "description",
    )
    list_display = (
        "code",
        "description",
        "valid_from",
        "valid_to",
    )


@admin.register(models.ReceiptType)
class ReceiptTypeAdmin(admin.ModelAdmin):
    search_fields = (
        "code",
        "description",
    )
    list_display = (
        "code",
        "description",
        "valid_from",
        "valid_to",
    )


class ReceiptHasFileFilter(admin.SimpleListFilter):
    title = _("has file")
    parameter_name = "has_file"

    YES = "yes"
    NO = "no"

    def lookups(
        self,
        request: HttpRequest,
        model_admin: admin.ModelAdmin,
    ) -> Sequence[tuple[str, str]]:
        return (
            (self.YES, _("Yes")),
            (self.NO, _("No")),
        )

    def queryset(
        self,
        request: HttpRequest,
        queryset: QuerySet[models.ReceiptPDF],
    ) -> QuerySet[models.ReceiptPDF]:
        if self.value() == self.YES:
            return queryset.exclude(pdf_file="")
        if self.value() == self.NO:
            return queryset.filter(pdf_file="")
        return queryset


@admin.register(models.ReceiptPDF)
class ReceiptPDFAdmin(admin.ModelAdmin):
    list_display = (
        "receipt_id",
        "taxpayer",
        "receipt",
        "client_name",
        "has_file",
    )
    list_filter = (
        ReceiptHasFileFilter,
        "receipt__point_of_sales__owner",
        "receipt__receipt_type",
    )
    raw_id_fields = ("receipt",)

    def get_queryset(self, request: HttpRequest) -> QuerySet[models.ReceiptPDF]:
        return (
            super()
            .get_queryset(request)
            .select_related(
                "receipt",
                "receipt__point_of_sales__owner",
                "receipt__receipt_type",
            )
        )

    @admin.display(description=_("taxpayer"))
    def taxpayer(self, obj: models.ReceiptPDF) -> str:
        return str(obj.receipt.point_of_sales.owner)

    @admin.display(boolean=True, description=_("Has file"), ordering="pdf_file")
    def has_file(self, obj: models.ReceiptPDF) -> bool:
        return bool(obj.pdf_file)

    @admin.action(description=_("Generate pdf"))
    def generate_pdf(
        self,
        request: HttpRequest,
        queryset: QuerySet[models.ReceiptPDF],
    ) -> None:
        for pdf in queryset:
            pdf.save_pdf()

    actions = (generate_pdf,)


@admin.register(models.ReceiptValidation)
class ReceiptValidationAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "receipt_number",
        "successful",
        "cae",
        "processed_date",
    )

    raw_id_fields = ("receipt",)

    @admin.display(description=_("receipt number"), ordering="receipt_id")
    def receipt_number(self, obj: models.ReceiptValidation) -> str | None:
        return obj.receipt.formatted_number

    @admin.display(boolean=True, description=_("result"), ordering="result")
    def successful(self, obj: models.ReceiptValidation) -> bool:
        return obj.result == models.ReceiptValidation.RESULT_APPROVED


admin.site.register(models.ConceptType)
admin.site.register(models.DocumentType)
admin.site.register(models.VatType)
admin.site.register(models.TaxType)
admin.site.register(models.Observation)



from __future__ import annotations

from unittest.mock import MagicMock
from unittest.mock import patch

import pytest

from django_afip import models
from django_afip.factories import ReceiptFactory


@pytest.fixture
def disable_durability_check() -> None:
    """Disable the global fixture of the same name."""


@pytest.mark.django_db
def test_raises() -> None:
    """Calling ``validate`` inside a transaction should raise."""

    receipt = ReceiptFactory()
    queryset = models.Receipt.objects.filter(pk=receipt.pk)
    ticket = MagicMock()

    with (
        patch(
            "django_afip.models.ReceiptQuerySet._assign_numbers",
            spec=True,
        ) as mocked_assign_numbers,
        pytest.raises(RuntimeError),
    ):
        # TYPING: django-stubs can't handle methods in querysets
        queryset.validate(ticket)  # type: ignore[attr-defined]

    assert mocked_assign_numbers.call_count == 0



from __future__ import annotations

import warnings

from django.utils.functional import cached_property
from django_renderpdf.views import PDFView

from django_afip import models
from django_afip.pdf import PdfBuilder


class ReceiptPDFView(PDFView):
    #: The PDF Builder class to use for generating PDF files.
    #:
    #: Set this to a custom subclass if you need custom behaviour for your PDF files.
    builder_class = PdfBuilder

    @cached_property
    def receipt(self) -> models.Receipt:
        """Returns the receipt.

        Returns the same in-memory instance during the whole request."""

        return models.Receipt.objects.select_related(
            "receipt_type",
            "point_of_sales",
        ).get(
            pk=self.kwargs["pk"],
        )

    @cached_property
    def builder(self) -> PdfBuilder:
        """Returns the pdf builder.

        Returns the same in-memory instance during the whole request."""

        return self.builder_class()

    @property
    def download_name(self) -> str:
        """Return the filename to be used when downloading this receipt."""
        return f"{self.receipt.formatted_number}.pdf"

    def get_template_names(self) -> list[str]:
        """Return the templates use to render the Receipt PDF.

        See :meth:`~.PdfBuilder.get_template_names` for exact implementation details.
        """
        return self.builder.get_template_names(self.receipt)

    @staticmethod
    def get_context_for_pk(pk: int, *args, **kwargs) -> dict:
        """Returns the context for a receipt.

        Note that this uses ``PdfBuilder`` and not ``self.builder_class`` due to legacy
        reasons.

        .. deprecated:: 12.0

            This method is deprecated, use :meth:`~.PdfBuilder.get_context` instead.
        """
        warnings.warn(
            "ReceiptPDFView.get_context_for_pk is deprecated; "
            "use PdfBuilder.get_context instead",
            DeprecationWarning,
            stacklevel=2,
        )
        receipt = models.Receipt.objects.get(pk=pk)
        return PdfBuilder().get_context(receipt)

    def get_context_data(self, pk: int, **kwargs) -> dict:
        context = super().get_context_data(pk=pk, **kwargs)
        context.update(self.builder.get_context(self.receipt))
        return context



# Generated by Django 4.2.5 on 2023-09-22 20:38
from __future__ import annotations

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("afip", "0012_optionaltype_optional_alter_code_in_generics"),
    ]

    operations = [
        migrations.AlterField(
            model_name="receiptentry",
            name="quantity",
            field=models.DecimalField(
                decimal_places=2, max_digits=15, verbose_name="quantity"
            ),
        ),
    ]



# Generated by Django 5.1.1 on 2024-10-03 16:12
from __future__ import annotations

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("afip", "0014_alter_pointofsales_blocked_alter_taxpayer_logo"),
    ]

    operations = [
        migrations.AlterField(
            model_name="taxpayer",
            name="logo",
            field=models.ImageField(
                blank=True,
                help_text="A logo to use when generating printable receipts.",
                null=True,
                upload_to="afip/taxpayers/logos/",
                verbose_name="logo",
            ),
        ),
    ]
